/*
  Macro           eHelp_SyntaxCheck
  Author          Carlo Hogeveen
  Website         ecarlo.nl/tse
  Compatibility   TSE v4 upwards, all TSE variants (intended, not tested yet)
  Version         v0.22   23 Mar 2023

  This tool checks the syntax of the provided or current .ehelp file.

  Input:
    A .ehelp file as a macro parameter, or otherwise as the current file.

  Output
    A same-named .ehelp_syntaxcheck file at the same location.

    When and where appropriate it will contain errors, warnings and notes:
      Errors point out syntax that will cause eHelp to malfunction.
        For example, a link to a non-existing (sub)topic. A user selecting one
        would get no result.
      Warnings point out syntax that makes no sense, which will hamper
        maintaining the .ehelp file.
        For example, a subtopic that is not linked to. TSE's Help contains
        these. A Help user does not see them.
      Notes relay syntax check information.
        For example, statistics.

  Foreseen problem:
    The context facilitates, that a new syntax check can be started
    for a .ehelp file, while its previous syntax check is still running.
    This will typically happen if a syntax check is startied by saving a .ehelp
    file.
    In such a case the last syntax check must overrule all previous ones.
  Two independent solutions:
    The tool stops prematurely if the .ehelp file's file-time becomes greater
    than the tool's start-time.
    The tool builds the .ehelp_syntaxcheck file in memory, and then saves it:
    this will be close to a garantee that the last started one's report will
    prevail.


  HISTORY

  v0.1   14 Nov 2022
    Initial version.
    Only does a rough, high-level syntax check yet.

  v0.2   17 Nov 2022
    Added a superficial syntax check of the inner text of same-line tag pairs.
    It only checks which number of "^" and ";" is mandatory and allowed for the
    inner text depending on each tag name.

  v0.18   8 Mar 2023
    Just aligning the version number with eHelp.

*/







// Start of compatibility restrictions and mitigations


#ifdef LINUX
  #define WIN32 FALSE
#endif


#if WIN32
  // Window 32-bit TSE.
  integer REQUIRED_RUN_TIME_CENTISECONDS =   100 // Should be lowered later.
#else
  #ifdef LINUX
    // Linux TSE. (All versions are 32-bit.)
    integer REQUIRED_RUN_TIME_CENTISECONDS = 100
  #else
    // Windows and Dos 16-bit TSE. (Not supported by this macro.)
    integer REQUIRED_RUN_TIME_CENTISECONDS = 200
  #endif
#endif


#if WIN32

  dll "<Kernel32.dll>"

    integer proc Windows_FindFirstFile(
      string lpFileName:cstrval,
      string lpFindFileData:strptr
      ):"FindFirstFileA"

    //  integer proc Windows_FindNextFile(
    //    integer handle,
    //    string  lpFindFileData:strptr
    //    ):"FindNextFileA"

    integer proc Windows_FileTimeToSystemTime(
      string lpFileTime:strptr,
      string lpSystemTime:strptr
      ):"FileTimeToSystemTime"

    integer proc Windows_SystemTimeToTzSpecificLocalTime(
      integer lpTimeZoneInformation,
      string  lpUniversalTime:strptr,
      string  lpLocalTime:strptr
      ):"SystemTimeToTzSpecificLocalTime"

    integer proc Windows_FindClose(
      integer hFindFile
      ):"FindClose"
  end


  integer proc word_to_val(string word)
    integer result = Asc(word[2]) * 256 + Asc(word[1])
    return(result)
  end word_to_val


  string proc get_file_windows_yyyy_mm_dd_hh_mm_ss(string file_name)
    // A returned file_data structure can be longer than 255 bytes.
    // Just to be sure I reserved excessive overflow space.
    // For global variables space is reserved in the order of their definition.
    // For local  variables space is reserved in the reverse order.
    string  file_data_overflow_9 [MAXSTRINGLEN] = Format('':MAXSTRINGLEN)
    string  file_data_overflow_8 [MAXSTRINGLEN] = Format('':MAXSTRINGLEN)
    string  file_data_overflow_7 [MAXSTRINGLEN] = Format('':MAXSTRINGLEN)
    string  file_data_overflow_6 [MAXSTRINGLEN] = Format('':MAXSTRINGLEN)
    string  file_data_overflow_5 [MAXSTRINGLEN] = Format('':MAXSTRINGLEN)
    string  file_data_overflow_4 [MAXSTRINGLEN] = Format('':MAXSTRINGLEN)
    string  file_data_overflow_3 [MAXSTRINGLEN] = Format('':MAXSTRINGLEN)
    string  file_data_overflow_2 [MAXSTRINGLEN] = Format('':MAXSTRINGLEN)
    string  file_data_overflow_1 [MAXSTRINGLEN] = Format('':MAXSTRINGLEN)
    string  file_data            [MAXSTRINGLEN] = Format('':MAXSTRINGLEN)
    integer file_handle                         = -1
    string  file_last_write_time            [8] = Format('':8)
    string  local_time           [MAXSTRINGLEN] = Format('':MAXSTRINGLEN)
    string  system_time          [MAXSTRINGLEN] = Format('':MAXSTRINGLEN)
    string  yyyy_mm_dd_hh_mm_ss            [19] = ''

    file_handle = Windows_FindFirstFile(file_name, file_data)

    if file_handle <> -1
      file_last_write_time = file_data[21: 8]
      Windows_FileTimeToSystemTime(file_last_write_time, system_time)
      Windows_SystemTimeToTzSpecificLocalTime(0, system_time, local_time)

      yyyy_mm_dd_hh_mm_ss = Format(word_to_val(local_time[ 1:2]):4:'0', '-',
                                   word_to_val(local_time[ 3:2]):2:'0', '-',
                                   word_to_val(local_time[ 7:2]):2:'0'     ;

                                   word_to_val(local_time[ 9:2]):2:'0', ':',
                                   word_to_val(local_time[11:2]):2:'0', ':',
                                   word_to_val(local_time[13:2]):2:'0'     )
      Windows_FindClose(file_handle)
    endif

    // Silence the compiler about apparently unused overflow strings.
    if FALSE
      file_data_overflow_1 = file_data_overflow_1
      file_data_overflow_2 = file_data_overflow_2
      file_data_overflow_3 = file_data_overflow_3
      file_data_overflow_4 = file_data_overflow_4
      file_data_overflow_5 = file_data_overflow_5
      file_data_overflow_6 = file_data_overflow_6
      file_data_overflow_7 = file_data_overflow_7
      file_data_overflow_8 = file_data_overflow_8
      file_data_overflow_9 = file_data_overflow_9
    endif

    return(yyyy_mm_dd_hh_mm_ss)
  end get_file_windows_yyyy_mm_dd_hh_mm_ss

#endif


string proc get_file_yyyy_mm_dd_hh_mm_ss(string file_name)
  string yyyy_mm_dd_hh_mm_ss [19] = ''
  #if WIN32
    // Windows 32-bit TSE.
    // TSE rounds file times up to the next even second,
    // meaning the actual file time can be up to two seconds earlier.
    // Usually that is insignificant, but here that leads to flawed behaviour.
    // So in 32-bit Windows we use a Windows API instead of a TSE function.
    yyyy_mm_dd_hh_mm_ss = get_file_windows_yyyy_mm_dd_hh_mm_ss(file_name)
  #else
    #ifdef LINUX
      // Linux TSE.
      // File times have a 1 second precision.
      if FindThisFile(CurrFilename())
        // These functions need date time formats to be set elsewhere.
        yyyy_mm_dd_hh_mm_ss = FFDateStr() + ' ' + FFTimeStr()
      endif
    #else
      // Windows and Dos 16-bit TSE.
      // Not supported.
    #endif
  #endif
  return(yyyy_mm_dd_hh_mm_ss)
end get_file_yyyy_mm_dd_hh_mm_ss

// End of compatibility restrictions and mitigations.





// Global constants and semi-constants.
// Semi-constants are initialized once and then do not change.
string  CHECK_FQN             [MAXSTRINGLEN] = ''
integer CHECK_ID                             = 0
integer IS_BACKGROUND_PROCESS                = FALSE
string  MACRO_NAME            [MAXSTRINGLEN] = ''
integer MACRO_START_CENTISECONDS             = 0
string  MACRO_START_DATE_TIME [MAXSTRINGLEN] = ''



// Global variable
integer stack_level = 0



string proc de_regexp(string regexp)
  integer i                     = 1
  string  result [MAXSTRINGLEN] = regexp
  while i <= Length(result)
    if Pos(result[i], '.^$|?[]-~*+@#{}\')
      result = result[1 .. i - 1] + '\' + result[i .. MAXSTRINGLEN]
      i = i + 2
    else
      i = i + 1
    endif
  endwhile
  return(result)
end


integer proc wrn(string msg)
  if IS_BACKGROUND_PROCESS
    Message(msg)
  else
    Warn(msg)
  endif
  return(FALSE)
end wrn


integer proc err(string msg)
  AddLine(Format('Error at (', CurrLine(), ',', CurrPos(), '):'; msg), CHECK_ID)
  return(FALSE)
end err


// The is a helper function for yyyy_mm_dd_hh_mm_ss_subtract().

proc yyyy_mm_dd_hh_mm_ss_subtract_limits(    string  yyyy_mm_dd_hh_mm_ss,
                                             integer part_pos,
                                         var integer part_min,
                                         var integer part_full)
  integer year = 0

  part_min = 0

  case part_pos
    when 1
      part_full = 9999
    when 6
      part_min  = 1
      part_full = 12
    when 9
      part_min  = 1
      // For subtraction we need the max number of days of the previous month.
      case Val(yyyy_mm_dd_hh_mm_ss[6: 2])
        when 1, 2, 4, 6, 8, 9, 11
          part_full = 31
        when 3
          year = Val(yyyy_mm_dd_hh_mm_ss[1: 4])
          if year mod 4 == 0
            if year mod 100 == 0
              if year mod 400 == 0
                part_full = 29
              else
                part_full = 28
              endif
            else
              part_full = 29
            endif
          else
            part_full = 28
          endif
        when 2, 5, 10, 12
          part_full = 30
      endcase
    when 12
      part_full = 24
    when 15, 18
      part_full = 60
  endcase
end yyyy_mm_dd_hh_mm_ss_subtract_limits


// This function is flawed for subtractions of 1 month and more.
// You can only safely subtract one of these values or less:
//            0 years     (part_pos =  1)   (29 Feb 2024 - 1 year  fails)
//            0 months    (part_pos =  6)   (31 May      - 1 month fails)
//           28 days      (part_pos =  9)
//          672 hours     (part_pos = 12)
//       40,320 minutes   (part_pos = 15)
//    2,419,200 seconds   (part_pos = 18)
// For example, these subtractions work correctly:
//    2022-01-01 00:00:01 -  2 from part_pos 18 = 2021-12-31 23:59:59
//    2024-03-01 00:00:01 -  2 from part_pos 18 = 2024-02-29 23:59:59

proc yyyy_mm_dd_hh_mm_ss_subtract(var string  yyyy_mm_dd_hh_mm_ss_in_out,
                                      integer part_pos_in,
                                      integer subtract)
  integer ok                       = FALSE
  integer part_full                = 0
  integer part_length              = 0
  integer part_min                 = 0
  integer part_pos                 = part_pos_in
  integer part_value               = 0
  string  yyyy_mm_dd_hh_mm_ss [19] = yyyy_mm_dd_hh_mm_ss_in_out

  if  (part_pos in 1, 6, 9, 12, 15, 18)
  and isDigit(yyyy_mm_dd_hh_mm_ss[ 1: 4])
  and (   Val(yyyy_mm_dd_hh_mm_ss[ 6: 2]) in 1 .. 12)
  and (   Val(yyyy_mm_dd_hh_mm_ss[ 9: 2]) in 1 .. 31)
  and isDigit(yyyy_mm_dd_hh_mm_ss[12: 2])
  and (   Val(yyyy_mm_dd_hh_mm_ss[12: 2]) in 0 .. 23)
  and isDigit(yyyy_mm_dd_hh_mm_ss[15: 2])
  and (   Val(yyyy_mm_dd_hh_mm_ss[15: 2]) in 0 .. 59)
  and isDigit(yyyy_mm_dd_hh_mm_ss[18: 2])
  and (   Val(yyyy_mm_dd_hh_mm_ss[18: 2]) in 0 .. 59)
    ok = TRUE
  endif

  while ok
  and   subtract
  and   part_pos
    part_length = iif(part_pos == 1, 4, 2)
    yyyy_mm_dd_hh_mm_ss_subtract_limits(yyyy_mm_dd_hh_mm_ss,
                                        part_pos,
                                        part_min,
                                        part_full)
    if part_full
      part_value  = Val(yyyy_mm_dd_hh_mm_ss[part_pos: part_length])
      part_value  = part_value - subtract
      subtract    = 0
      while part_value < part_min
        part_value = part_value + part_full
        subtract   = subtract + 1
      endwhile
      yyyy_mm_dd_hh_mm_ss[part_pos: part_length] =
        Format(part_value: part_length: '0')
      part_pos = iif(part_pos <= 6, 1, part_pos - 3)
    else
      ok = FALSE
    endif
  endwhile

  if ok
    yyyy_mm_dd_hh_mm_ss_in_out = yyyy_mm_dd_hh_mm_ss
  endif
end yyyy_mm_dd_hh_mm_ss_subtract


integer proc i_am_still_the_last_ehelp_syntaxcheck()
  string  ehelp_file_date_time [19] = ''
  integer result                    = TRUE

  ehelp_file_date_time = get_file_yyyy_mm_dd_hh_mm_ss(CurrFilename())
  if FindThisFile(CurrFilename())
    // TSE rounds up file date-times to the next even second.
    // Compensate for that by subtracting 2 seconds.
    yyyy_mm_dd_hh_mm_ss_subtract(ehelp_file_date_time, 18, 2)
    result = (MACRO_START_DATE_TIME >= ehelp_file_date_time)
  endif
  return(result)
end i_am_still_the_last_ehelp_syntaxcheck


integer proc get_top_line()
  return(GetBufferInt(MACRO_NAME + ':line_at_level_' + Str(stack_level)))
end get_top_line


string proc get_top_tag()
  return(GetBufferStr(MACRO_NAME +  ':tag_at_level_' + Str(stack_level)))
end get_top_tag


integer proc push_tag(string tag_name)
  integer ok = FALSE
  stack_level = stack_level + 1
  if  SetBufferInt(MACRO_NAME + ':line_at_level_' + Str(stack_level),
                   CurrLine())
  and SetBufferStr(MACRO_NAME +  ':tag_at_level_' + Str(stack_level),
                   tag_name)
    ok = TRUE
  endif
  return(ok)
end push_tag


integer proc pop_tag()
  integer ok                      = FALSE
  integer tag_line                = 0
  string  tag_name [MAXSTRINGLEN] = ''
  if stack_level > 0
    tag_line = get_top_line()
    tag_name = get_top_tag()
    DelBufferVar(MACRO_NAME + ':line_at_level_' + Str(stack_level))
    DelBufferVar(MACRO_NAME +  ':tag_at_level_' + Str(stack_level))
    stack_level = stack_level - 1
    // The following tags must open and close in the same line.
    if (tag_name in 'BOLD',
                    'GENERATED_DATE_TIME',
                    'GENERATOR_TOOL',
                    'GENERATOR_VERSION',
                    'INFO',
                    'ITALIC',
                    'LINK',
                    'REDIRECT',
                    'SUBTOPIC',
                    'TOPIC')
      if tag_line == CurrLine()
        ok = TRUE
      else
        ok = err('Tag </' + tag_name +
                 '> not closes in same line as opening tag.')
      endif
    else
      ok = TRUE
    endif
  endif
  return(ok)
end pop_tag


integer proc has_ancestor(string tag)
  integer i      = 0
  integer result = FALSE
  for i = stack_level downto 1
    if GetBufferStr(MACRO_NAME + ':tag_at_level_' + Str(i)) == tag
      result = TRUE
      break
    endif
  endfor
  return(result)
end has_ancestor


integer proc check_tag_hierarchy_and_sameline()
  string  closing_tag [MAXSTRINGLEN] = ''
  string  found_tag   [MAXSTRINGLEN] = ''
  integer ok                         = TRUE
  string  opening_tag [MAXSTRINGLEN] = ''
  string  search_options         [3] = 'gix'
  string  top_tag     [MAXSTRINGLEN] = ''

  PushLocation()
  while ok
  and   lFind('<[~<]+>', search_options)
    found_tag = GetFoundText()
    top_tag   = get_top_tag()
    if found_tag[2] == '/'
      opening_tag = ''
      closing_tag = Upper(found_tag[3: Length(found_tag) - 3])
      if top_tag == 'HELP'
        // Within HELP tags almost all other tags are ignored.
        if (closing_tag in 'BOLD', 'INFO', 'ITALIC', 'LINK', 'SUBTOPIC', 'HELP')
          if closing_tag == top_tag
            ok = pop_tag()
          endif
        endif
      elseif top_tag == 'TEXT'
        // After opening TEXT tag, anything but closing TEXT tag is ignored.
        if closing_tag == 'TEXT'
          ok = pop_tag()
        endif
      else
        // Note that what should be inner HELP tags, need to be checked here as
        // well, namely for being closed without having been opened.
        if (closing_tag in 'BOLD', 'EHELP', 'GENERATED_DATE_TIME',
                           'GENERATOR_TOOL', 'GENERATOR_VERSION', 'HELP',
                           'INFO', 'ITALIC', 'LINK', 'META', 'REDIRECT',
                           'REDIRECTS', 'SUBTOPIC', 'TEXT', 'TOPIC', 'TOPICS')
          if top_tag == closing_tag
            ok = pop_tag()
          elseif not has_ancestor('HELP')
            ok = err('Cannot close tag </' + closing_tag + '> while <' +
                            top_tag + '> is open.')
          endif
        elseif not has_ancestor('HELP')
            ok = err('Unknown closing tag </' + closing_tag + '>.')
        endif
      endif
    else
      opening_tag = Upper(found_tag[2: Length(found_tag) - 2])
      closing_tag = ''
      if top_tag == 'HELP'
        // Within HELP tags all its outer tags are treated as text.
        if (opening_tag in 'BOLD', 'INFO', 'ITALIC', 'LINK', 'SUBTOPIC')
          push_tag(opening_tag)
        elseif opening_tag == 'HELP'
          ok = err('<HELP> cannot be a child tag of itself.')
        endif
      elseif top_tag == 'TEXT'
        // Within TEXT tags all other tags are treated as text.
        NoOp()
      else
        case opening_tag
          when 'INFO', 'LINK', 'SUBTOPIC'
            if (top_tag in 'BOLD', 'ITALIC')
              push_tag(opening_tag)
            else
              ok = err('<' + opening_tag + '> cannot be a child tag of <' +
                       top_tag + '>.')
            endif
          when 'BOLD', 'ITALIC'
              ok = err('<' + opening_tag + '> cannot be a child tag of <' +
                       top_tag + '>.')
          when 'META', 'REDIRECTS', 'TOPICS'
            if top_tag == 'EHELP'
              push_tag(opening_tag)
            else
              ok = err('<' + opening_tag +
                       '> must be a child tag of <EHELP>.')
            endif
          when 'HELP', 'TOPIC'
            if top_tag == 'TOPICS'
              push_tag(opening_tag)
            else
              ok = err('<' + opening_tag +
                       '> must be a child tag of <TOPICS>.')
            endif
          when 'REDIRECT'
            if top_tag == 'REDIRECTS'
              push_tag(opening_tag)
            else
              ok = err('<REDIRECT> must be a child tag of <REDIRECTS>.')
            endif
          when 'TEXT',
               'GENERATOR_TOOL', 'GENERATOR_VERSION', 'GENERATED_DATE_TIME'
            if top_tag == 'META'
              push_tag(opening_tag)
            else
              ok = err('<' + opening_tag + '> must be a child tag of <META>.')
            endif
          when 'EHELP'
            if top_tag == ''
              push_tag(opening_tag)
            else
              ok = err('<EHELP> must be the top tag.')
            endif
          otherwise
            if not has_ancestor('HELP')
              ok = err('Unknown opening tag <' + opening_tag + '>.')
            endif
        endcase
      endif
    endif
    search_options = 'ix+'
  endwhile
  PopLocation()
  return(ok)
end check_tag_hierarchy_and_sameline


integer proc check_tags_aloneinline()
  integer both_tags_for_whole_line   = FALSE
  string  found_tag   [MAXSTRINGLEN] = ''
  integer in_help                    = FALSE
  integer in_text                    = FALSE
  integer is_closing_tag             = FALSE
  integer ok                         = TRUE
  string  search_options         [3] = 'gix'
  integer single_tag_for_whole_line  = FALSE
  integer tag_slicer                 = 0
  string  tag_name    [MAXSTRINGLEN] = ''

  PushLocation()

  while ok
  and   lFind('<[~<]+>', search_options)
    found_tag      = GetFoundText()
    is_closing_tag = (found_tag[2] == '/')
    tag_slicer     = iif(is_closing_tag, 3, 2)
    tag_name       = Upper(found_tag[tag_slicer: Length(found_tag) - tag_slicer])

    single_tag_for_whole_line = FALSE
    both_tags_for_whole_line  = FALSE

    if     in_help
      if  is_closing_tag
      and tag_name == 'HELP'
        in_help                   = FALSE
        single_tag_for_whole_line = TRUE
      endif
    elseif in_text
      if  is_closing_tag
      and tag_name == 'TEXT'
        in_text                   = FALSE
        single_tag_for_whole_line = TRUE
      endif
    else
      case tag_name
        when 'EHELP',
             'META',
             'REDIRECTS',
             'TOPICS'
          single_tag_for_whole_line = TRUE
        when 'GENERATED_DATE_TIME',
             'GENERATOR_TOOL',
             'GENERATOR_VERSION',
             'REDIRECTS',
             'TOPIC'
          both_tags_for_whole_line  = TRUE
        when 'HELP'
          in_help                   = TRUE
          single_tag_for_whole_line = TRUE
        when 'TEXT'
          in_text                   = TRUE
          single_tag_for_whole_line = TRUE
      endcase
    endif

    if single_tag_for_whole_line
      if lFind('^ *' + de_regexp(found_tag) + ' *$', 'cgix')
        EndLine()
      else
        ok = err('Tag ' + found_tag + ' must be alone in its line.')
      endif
    endif

    if both_tags_for_whole_line
      if lFind('^ *<' + de_regexp(tag_name) + '>[~<]*</' +
               de_regexp(tag_name) + '> *$', 'cgix')
        EndLine()
      else
        ok = err('Tags for ' + tag_name + ' must be alone in their line.')
      endif
    endif

    search_options = 'ix+'
  endwhile

  PopLocation()
  return(ok)
end check_tags_aloneinline


integer proc check_same_line_tag_pair_inner_syntax()
  //  This is a quick superficial syntax check for "^" and ";" usage.
  //
  //  Tags that must pair in the same line:
  //    Such tags that must be syntax checked:
  //      <INFO>
  //      <LINK>
  //      <REDIRECT>
  //      <SUBTOPIC>
  //      <TOPIC>
  //    Such tags that need no syntax check:
  //      <BOLD>
  //      <GENERATED_DATE_TIME>
  //      <GENERATOR_TOOL>
  //      <GENERATOR_VERSION>
  //      <ITALIC>
  string  found_tag      [MAXSTRINGLEN] = ''
  string  inner_text     [MAXSTRINGLEN] = ''
  integer inner_text_from               = 0
  integer inner_text_length             = 0
  integer inner_text_to                 = 0
  integer ok                            = TRUE
  string  search_options            [2] = 'gx'
  string  source_address [MAXSTRINGLEN] = ''
  string  tag_name       [MAXSTRINGLEN] = ''
  string  target_address [MAXSTRINGLEN] = ''

  while lFind('<[~/<\\]+>', search_options)
    found_tag = GetFoundText()
    if (Upper(found_tag) in '<INFO>', '<LINK>', '<REDIRECT>', '<SUBTOPIC>',
                            '<TOPIC>')
      tag_name        = Upper(found_tag[2: Length(found_tag) - 2])
      inner_text_from = CurrPos() + Length(found_tag)
      if lFind('</' + tag_name + '>', 'ci')
        inner_text_to     = CurrPos() - 1
        inner_text_length = inner_text_to - inner_text_from + 1
        inner_text        = GetText(inner_text_from, inner_text_length)
        case NumTokens(inner_text, '^')
          when 1
            case tag_name
              when 'INFO', 'SUBTOPIC', 'TOPIC'
                if Pos(';', inner_text)
                  err('Between ' + tag_name + ' tags a ";" is not allowed.')
                endif
              when 'LINK'
                if NumTokens(inner_text, ';') > 2
                  err('Between LINK tags only one ";" is allowed.')
                endif
              when 'REDIRECT'
                err('REDIRECT tags need "^" between source and target address.')
            endcase
          when 2
            source_address = ''
            target_address = ''
            case tag_name
              when 'LINK'
                target_address = GetToken(inner_text, '^', 1)
              when 'REDIRECT'
                source_address = GetToken(inner_text, '^', 1)
                target_address = inner_text[Length(source_address) + 2: MAXSTRINGLEN]
              otherwise
                ok = err('Between ' + tag_name + ' tags "^" is not allowed.')
            endcase
            if source_address <> ''
              if NumTokens(source_address, ';') > 2
                ok = err('Source address between REDIRECT tags has more than one ";".')
              endif
            endif
            if target_address <> ''
              if NumTokens(target_address, ';') > 2
                ok = err('Target address between ' + tag_name + ' tags has more than one ";".')
              endif
            endif
          otherwise
            if (tag_name in 'LINK', 'REDIRECT')
              ok = err('Between ' + tag_name + ' tags only one "^" is allowed.')
            else
              ok = err('Between ' + tag_name + ' tags no "^" are allowed.')
            endif
        endcase
      endif
    endif
    search_options = 'x+'
  endwhile

  return (ok)
end check_same_line_tag_pair_inner_syntax


proc Main()
  string  dos_cmd   [MAXSTRINGLEN] = Query(DosCmdLine)
  string  ehelp_fqn [MAXSTRINGLEN] = ''
  string  mac_cmd   [MAXSTRINGLEN] = Query(MacroCmdLine)
  integer ok                       = TRUE
  integer old_DateFormat           = Set(DateFormat, 6)
  integer old_TimeFormat           = Set(TimeFormat, 1)
  integer org_id                   = GetBufferId()

  MACRO_NAME               = SplitPath(CurrMacroFilename(), _NAME_)
  MACRO_START_CENTISECONDS = GetTime()
  MACRO_START_DATE_TIME    = GetDateStr() + ' ' + GetTimeStr()

  // '-e ' is intentionally not tested for, to be available for test purposes.
  // Because it is eHelp that starts TSE from the command line with this macro
  // with '-e' as a background process, '-e ' should not occur outside testing.
  if Pos('-e'  + Lower(MACRO_NAME), Lower(dos_cmd))
    IS_BACKGROUND_PROCESS = TRUE
  endif

  if mac_cmd <> ''
    ehelp_fqn = ExpandPath(GetFileToken(mac_cmd, 1))
    if Lower(SplitPath(ehelp_fqn, _EXT_)) <> '.ehelp'
      wrn('Macro parameter is not a .ehelp filename.')
      ok = FALSE
    elseif not EquiStr(ehelp_fqn, CurrFilename())
      if not EditThisFile(ehelp_fqn, _DONT_PROMPT_)
        wrn("Could not open macro parameter's .ehelp file.")
        ok = FALSE
      endif
    endif
  endif

  if  ok
  and CurrExt() <> '.ehelp'
    wrn('Current file is not a .ehelp file.')
    ok = FALSE
  endif

  if  ok
  and (      FileChanged()
      or not FileExists(CurrFilename()))
    wrn('Current .ehelp buffer is changed or not saved as a file.')
    ok = FALSE
  endif

  if ok
    CHECK_FQN = SplitPath(CurrFilename(), _DRIVE_|_PATH_|_NAME_) +
                '.ehelp_syntaxcheck'
    CHECK_ID  = GetBufferId(CHECK_FQN)
    if CHECK_ID
      EmptyBuffer(CHECK_ID)
    else
      CHECK_ID = CreateTempBuffer()
      ChangeCurrFilename(CHECK_FQN, _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)
      GotoBufferId(org_id)
    endif
  endif

  // Basic, fast syntax check parts.
  ok = ok and check_tag_hierarchy_and_sameline()
  ok = ok and check_tags_aloneinline()
  ok = ok and check_same_line_tag_pair_inner_syntax()

  // Add an if-block like this for each slow syntax checks part.
  if  ok
  and i_am_still_the_last_ehelp_syntaxcheck()
    // No slow syntax checks yet.
  endif

  GotoBufferId(org_id)
  Set(DateFormat, old_DateFormat)
  Set(TimeFormat, old_TimeFormat)

  if CHECK_ID
    // In Linux the .ehelp_syntaxcheck file must become at least 1 second newer
    // than its old time, so that eHelp can be determine when it is updated.
    // So in Linux this tool makes sure it is at least a second after its start
    // time before it updates the .ehelp_syntaxcheck file.
    // (In Windows its file time can be determined to the milisecond.)
    while Abs(GetTime() - MACRO_START_CENTISECONDS) < REQUIRED_RUN_TIME_CENTISECONDS
      Delay(1)
    endwhile

    if i_am_still_the_last_ehelp_syntaxcheck()
      GotoBufferId(CHECK_ID)
      SaveAs(CHECK_FQN, _DONT_PROMPT_|_OVERWRITE_)
      GotoBufferId(org_id)

      if not IS_BACKGROUND_PROCESS
        MsgBoxBuff(CHECK_FQN, _OK_, CHECK_ID)
      endif
    endif
  endif

  AbandonFile(CHECK_ID)
  PurgeMacro(MACRO_NAME)
  if IS_BACKGROUND_PROCESS
    AbandonEditor()
  endif
end Main

