/*
  Macro           helphelp
  Author          Carlo Hogeveen
  Website         eCarlo.nl/tse
  Compatibility   TSE v4 upwards: GUI, Console, Linux
  Version         0.6   23 Dec 2022


  This tool manages TSE Help systems.
  TSE's own built-in Help system is always the fall-back.

  Note:
    This is a test version of helphelp.
    One of its purposes is to find out if and how well the two non-TSE Help
    systems can coexist.


  INSTALLATION

  Copy this file into TSE's "mac" directory and compile it there,
  for example by opening the file in TSE and applying the Macro Compile menu.

  Optionally execute "helphelp" as a macro to configure TSE Help systems
  other than eHelp.
  For example, use the TSE menu Macro, Execute, helphelp.

  Warning:
    If GetHelp is loaded, it overrules the keys <F1>, <Ctrl F1>, <Shift F1> and
    <CtrlAlt H>, which will then use GetHelp's help directly, regardless of
    helphelp's configuration.


  DOCUMENTATION

  This tool switches TSE between Help systems:

    Semware's own Help    from 2005.
      This is the default help system, and also the fall-back help system if
      one of the other ones cannot find a help topic.

      It has been updated with tidbits since 2005, but almost all of TSE's
      changes since 2005 are not documented in its Help, and are instead
      briefly mentioned in its read.me file.

    Carlo's       eHelp   from 2022.
      It retrieves its Help topics from configurable .ehelp text files:
        tsehelp.ehelp
          A copy of TSE's built-in Help that is auto-generated by eHelp.
        tsehelp_eCarlo.ehelp
          One day will contain improved, up-to-date help for TSE.
        tsehelp_<name>.ehelp
          Help file by user <name> or for tool <name> or both.
          You can create or receive and configure as many as you like.

      The tool is in development and tsehelp_eCarlo.ehelp file is still a test
      file.

    Chris Antos'  GetHelp from 2002.
      GetHelp intends to improve TSE's Help and to provide access to non-TSE
      help systems from TSE, optionally based on a file's extension.

      GetHelp is included here because it exists.
      Its bugs and help files have not been maintained since 2002.
      AFAIK, based on mailing list info, no one uses it as a TSE Help system.
      It is used by a few of Chris' tools that do have users.
      These tools specifically use it to display their own help file,
      and then the above warning about <F1>, etc. key take-over applies.

      I am not a GetHelp user, and will not provide support for GetHelp,
      only for this helphelp interface to it.


  Acknowledgement
    This tool is partially based on and backwards compatibel with Chris Antos'
    version of the HELPHELP macro, published 19 May 2002 at https://semware.com
    as part of gethelp40.zip.

    I learned a lot of TSE's Help's technical information from Chris'
    gethelp40.zip package.


  HISTORY

  v0.5   14 Dec 2022
    Is now fully configurable, and in theory backwards compatible with GetHelp.

  v0.6   23 Dec 2022
    This macro now maintains a string buffer variable
    "helphelp:ExternalHelpSystem" for the "help topic buffer", buffer 8,
    containing which external help system successfully provided the topic's
    content, or an empty string if none did.

    eHelp needs this to know, whether it is still in charge, and therefore is
    allowed to switch between .ehelp files.

*/





/*
  T E C H N I C A L   B A C K G R O U N D   I N F O R M A T I O N


  TSE has built-in knowledge of a macro named "helphelp" that can overrule and
  augment TSE's help for a requested help topic.
  If at TSE's start-up "helphelp.mac" exists on disk in TSE's search path,
  then TSE calls it for every help request during the TSE session.


  INPUT

  A user can call helphelp with no parameters to configure it.


  TSE's Help system makes the help topic buffer (buffer id == 8) the current
  buffer, and then calls helphelp with one of these MacroCmdLine parameter
  formats:
    -r <help entry>
    -h <not a help entry>

  Parameter "-r <help entry>" is a request to retrieve the text of the help
  topic that is identified by <help entry>.
  In this case <help entry> can be a topic, a subtopic in a topic, or a help
  entry that is redirected to either.

  Parameter "-h <not a help entry>" is a request for an error text that is
  appropriate when halting the help request for <not a help entry>.
  TSE itself sends a "-h" (I suspect meaning "halt") request when it encounters
  a help error situation.
  systems.
  eHelp currently declines to provide such an error text.
  GetHelp treats it like an ordinarry help request.
  TSE itself, the default and fall-back help system, usually returns:
    Topic '<not a help entry>' not found


  PROCESS

  Default helphelp will pass the request by MacroCmdLine as-is to each of its
  configured Help systems. With the exception of GetHelp, which wants to be
  called with "-1" (minus one) instead of "-r" and "-h".

  helphelp expects to receive a MacroCmdLine, for which Val(Query(MacroCmdLine))
  is truthy or falsy. (In TSE falsy is 0 and truthy is any other value.)

  If helphelp receives falsy from a help system, then it tries the next
  configured help system, and if all such return falsy, then helphelp returns
  the MacroCmdLine string "false" to TSE's Help system.

  If helphelp receives truthy from a help system, then it stops trying other
  help systems, and returns the MacroCmdLine string "true" to TSE.


  OUTPUT

  If helphelp returns "true", then TSE's Help system expects its initial
  current buffer, which is the help topic buffer with id 8, to be filled with
  help text in _DISPLAY_HELP_ format.

  The latter for example means with B and /B around bold text.

  Also if helphelp returns "true", the string buffer variable
  "helphelp:ExternalHelpSystem" for the "help topic buffer", buffer 8,
  is set, so that external help systems can detect when TSE presents their
  help, and then can provide functionality to that help, like extra keys.

*/





// Start of compatibility restrictions and mitigations


#ifdef LINUX
  #define WIN32 FALSE
#endif


#ifdef WIN32
#else
   16-bit versions of TSE are not supported. You need at least TSE 4.0.
#endif

#ifdef EDITOR_VERSION
#else
   Editor Version is older than TSE 3.0. You need at least TSE 4.0.
#endif

#if EDITOR_VERSION < 4000h
   Editor Version is older than TSE 4.0. You need at least TSE 4.0.
#endif


#ifndef INTERNAL_VERSION
  #define INTERNAL_VERSION 0
#endif


#ifdef LINUX
  string SLASH [1] = '/'
#else
  string SLASH [1] = '\'
#endif

// End of compatibility restrictions and mitigations





// Global constants and semi-constants
#define CHANGECURRFILENAME_OPTIONS      _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_
string  HELPHELP_CFG_LIST_FOOTER [75] = '{Ins}-Add {Del}-Delete {Ctrl Up/Down}-HelpSystem Up/Down {Escape}-Save&Exit'
integer HELPHELP_CFG_LIST_MIN_WIDTH   = 0
string  HELPHELP_CFG_LIST_TITLE  [58] = 'List of configured help systems, as will be used top-down '
#define HELPHELP_SYSTEM_ADD             1
#define HELPHELP_SYSTEM_DEL             2
#define HELPHELP_SYSTEM_DOWN            3
#define HELPHELP_SYSTEM_UP              4
integer HELP_SYSTEMS_ID               = 0
#define HELP_TOPIC_ID                   8
string  MACRO_NAME     [MAXSTRINGLEN] = ''

#ifdef LINUX
  string  DEFAULT_HELP_SYSTEMS     [19] = 'eHelp.mac ehelp.mac'
#else
  string  DEFAULT_HELP_SYSTEMS      [9] = 'eHelp.mac'
#endif



// Global variables
integer helphelp_in_progress = FALSE
integer helphelp_enabled     = TRUE
integer list_action          = _NONE_
integer num_help_systems     = 0


/*
  Give a warning if WriteProfileStr() or WriteProfileInt() fails.
*/
integer proc write_profile_str(string section_name,
                               string item_name,
                               string item_value)
  integer result = WriteProfileStr(section_name,
                                   item_name,
                                   item_value)
  if not result
    Warn('Failed to write'; MACRO_NAME; 'configuration to tse.ini file.')
  endif
  return(result)
end write_profile_str


integer proc write_profile_int(string  section_name,
                               string  item_name,
                               integer item_value)
  integer result = WriteProfileInt(section_name,
                                   item_name,
                                   item_value)
  if not result
    Warn('Failed to write'; MACRO_NAME; 'configuration to tse.ini file.')
  endif
  return(result)
end write_profile_int


/*
  search_path works like SearchPath, but without looking in the current
  directory too if subdir is supplied.
  For example, if subdir is supplied, and paths is "path1;path2", then these
  directories are searched in this order by search_path:
    path1
    path1\subdir
    path2
    path2\subdir
    LoadDir()
    LoadDir()\subdir
    SplitPath(LoadDir(TRUE), _DRIVE_|_PATH_)
    SplitPath(LoadDir(TRUE), _DRIVE_|_PATH_)\subdir

  search_path_helper() should not be called directly.
  It is a helper proc for search_path().
*/
string proc search_path_helper(string filename, string path , string subdir)
  string result [MAXSTRINGLEN] = ''

  if FileExists(path + filename)
    result = path + filename
  elseif subdir <> ''
  and    FileExists(path + subdir + SLASH + filename)
    result = path + subdir + SLASH + filename
  endif

  return(result)
end search_path_helper


string proc search_path(string filename, string paths, string subdir)
  integer i                          = 0
  string  path        [MAXSTRINGLEN] = ''
  string  program_dir [MAXSTRINGLEN] = ''
  string  result      [MAXSTRINGLEN] = ''

  for i = 1 to NumTokens(paths, ';')
    path   = GetToken(paths, ';', i)
    result = search_path_helper(filename, path + SLASH, subdir)
    if result <> ''
      break
    endif
  endfor

  if result == ''
    path   = LoadDir()
    result = search_path_helper(filename, path, subdir)
  endif

  if result == ''
    program_dir = SplitPath(LoadDir(TRUE), _DRIVE_|_PATH_)
    if program_dir <> path
      result = search_path_helper(filename, program_dir, subdir)
    endif
  endif

  return(result)
end search_path


string proc get_helphelp_enabled()
  return(iif(helphelp_enabled, 'Enabled', 'Disabled'))
end get_helphelp_enabled


proc toggle_helphelp_enabled()
  helphelp_enabled = not helphelp_enabled
  write_profile_int(MACRO_NAME + ':Config', 'Enabled', helphelp_enabled)
end toggle_helphelp_enabled


proc initial_default_setup()
  string  help_system [MAXSTRINGLEN] = ''
  integer i                          = 0

  PushLocation()
  GotoBufferId(HELP_SYSTEMS_ID)
  if not NumLines()
    for i = 1 to NumTokens(DEFAULT_HELP_SYSTEMS, ' ')
      help_system = GetToken(DEFAULT_HELP_SYSTEMS, ' ', i)
      if not lFind(help_system, '^gi$')
      and    search_path(help_system, Query(TSEPath), 'mac') <> ''
        AddLine(help_system)
        write_profile_str(MACRO_NAME + ':HelpSystems', Str(i), help_system)
      endif
    endfor
  endif
  PopLocation()
end initial_default_setup


proc init_help_systems()
  string  item_name             [1] = '' // Not used.
  string  item_value [MAXSTRINGLEN] = ''
  PushLocation()
  GotoBufferId(HELP_SYSTEMS_ID)
  EmptyBuffer()
  num_help_systems = 0
  LoadProfileSection(MACRO_NAME + ':HelpSystems')
  while GetNextProfileItem(item_name, item_value)
    if search_path(item_value, Query(TSEPath), 'mac') == ''
      Warn(MACRO_NAME, ': configured help system "', item_value,
           '" does not exist as a .mac file.')
    else
      num_help_systems = num_help_systems + 1
      AddLine(SplitPath(item_value, _NAME_))
    endif
  endwhile
  PopLocation()
end init_help_systems


Keydef configure_systems_list_keys
  <Ins>                 list_action = HELPHELP_SYSTEM_ADD  PushKey(<Escape>)
  <GreyIns>             list_action = HELPHELP_SYSTEM_ADD  PushKey(<Escape>)
  <Del>                 list_action = HELPHELP_SYSTEM_DEL  PushKey(<Escape>)
  <GreyDel>             list_action = HELPHELP_SYSTEM_DEL  PushKey(<Escape>)
  <Ctrl CursorDown>     list_action = HELPHELP_SYSTEM_DOWN PushKey(<Escape>)
  <Ctrl GreyCursorDown> list_action = HELPHELP_SYSTEM_DOWN PushKey(<Escape>)
  <Ctrl CursorUp>       list_action = HELPHELP_SYSTEM_UP   PushKey(<Escape>)
  <Ctrl GreyCursorUp>   list_action = HELPHELP_SYSTEM_UP   PushKey(<Escape>)
  <Enter>     NoOp()
  <GreyEnter> NoOp()
end configure_systems_list_keys


proc cleanup_configure_systems_list()
  UnHook(cleanup_configure_systems_list)
  Disable(configure_systems_list_keys)
end cleanup_configure_systems_list


proc startup_configure_systems_list()
  UnHook(startup_configure_systems_list)
  Hook(_LIST_CLEANUP_, cleanup_configure_systems_list)
  list_action = _NONE_
  if Enable(configure_systems_list_keys)
    ListFooter(HELPHELP_CFG_LIST_FOOTER)
  endif
end startup_configure_systems_list


proc configure_help_systems()
  string  help_system [MAXSTRINGLEN] = ''
  integer i                          = 0
  string  mac_file    [MAXSTRINGLEN] = ''

  PushLocation()
  GotoBufferId(HELP_SYSTEMS_ID)
  if NumLines()
    PushBlock()
    MarkLine(1, NumLines())

    CreateTempBuffer()
    CopyBlock()
    PopBlock()
  else
    CreateTempBuffer()
  endif
  lReplace('.mac', '', 'gin$')
  BegFile()
  list_action = _NONE_
  repeat
    case list_action
      when HELPHELP_SYSTEM_ADD
        if Ask('New help system:', help_system)
          help_system = Trim(help_system)
          if      help_system <> ''
          and not lFind(help_system, '^gi$')
            if  search_path(help_system + '.mac', Query(TSEPath), 'mac') == ''
            and search_path(help_system + '.MAC', Query(TSEPath), 'mac') == ''
              Warn(MACRO_NAME, ': help system "', help_system,
                   '" does not exist as a .mac file.')
            else
              AddLine(help_system)
            endif
          endif
        endif
      when HELPHELP_SYSTEM_DEL
        KillLine()
      when HELPHELP_SYSTEM_DOWN
        if Down()
          SwapLines(CurrLine(), CurrLine() - 1)
        endif
      when HELPHELP_SYSTEM_UP
        if Up()
          SwapLines(CurrLine(), CurrLine() + 1)
        endif
    endcase
    Hook(_LIST_STARTUP_, startup_configure_systems_list)
  until not lList(HELPHELP_CFG_LIST_TITLE,
                  Max(LongestLineInBuffer(), HELPHELP_CFG_LIST_MIN_WIDTH),
                  NumLines(),
                  _ENABLE_HSCROLL_)
    and list_action == _NONE_
  if RemoveProfileSection(MACRO_NAME + ':HelpSystems')
    EmptyBuffer(HELP_SYSTEMS_ID)
    BegFile()
    for i = 1 to NumLines()
      GotoLine(i)
      help_system = Trim(GetText(1, MAXSTRINGLEN))
      mac_file    = search_path(help_system + '.mac', Query(TSEPath), 'mac')
      if mac_file == ''
        mac_file  = search_path(help_system + '.MAC', Query(TSEPath), 'mac')
      endif
      if mac_file == ''
        Warn(MACRO_NAME, ': just configured help system "', help_system,
             '" does not exist as a .mac file.')
      else
        help_system = SplitPath(mac_file, _NAME_|_EXT_)
        write_profile_str(MACRO_NAME + ':HelpSystems', Str(i), help_system)
      endif
    endfor
    init_help_systems()
  else
    Warn(MACRO_NAME, ': Unable to write new configuration settings to tse.ini.')
  endif
  PopLocation()
end configure_help_systems


integer proc get_num_help_systems()
  return(num_help_systems)
end get_num_help_systems


menu main_config_menu()
  title = 'External Help Systems Configuration'
  history

  "&External help systems for TSE"
    [get_helphelp_enabled(): 8],
    toggle_helphelp_enabled(),
    _MF_DONT_CLOSE_,
    'Toggle helphelp enabled/disabled'

  '&Which external help systems ...'
    [get_num_help_systems(): 8],
    configure_help_systems(),
    _MF_DONT_CLOSE_,
    'Configure which auxiliary help systems to use.'
end main_config_menu


proc configure_helphelp()
  main_config_menu()
end configure_helphelp


proc WhenPurged()
  AbandonFile(HELP_SYSTEMS_ID)
end WhenPurged


proc WhenLoaded()
  MACRO_NAME       = SplitPath(CurrMacroFilename(), _NAME_)
  helphelp_enabled = GetProfileInt(MACRO_NAME + ':Config', 'Enabled', TRUE)

  HELPHELP_CFG_LIST_MIN_WIDTH =
    Max(Length(       HELPHELP_CFG_LIST_TITLE ) + 5,
        Length(       HELPHELP_CFG_LIST_FOOTER) -
        StrCount('{', HELPHELP_CFG_LIST_FOOTER) -
        StrCount('}', HELPHELP_CFG_LIST_FOOTER))

  PushLocation()
  HELP_SYSTEMS_ID = CreateTempBuffer()
  ChangeCurrFilename(MACRO_NAME + ':HelpSystems', CHANGECURRFILENAME_OPTIONS)
  PopLocation()

  init_help_systems()
  initial_default_setup()
end WhenLoaded


proc Main()
  string  help_cmd_line  [MAXSTRINGLEN] = ''
  string  help_system    [MAXSTRINGLEN] = ''
  integer i                             = 0
  string  macro_cmd_line [MAXSTRINGLEN] = Query(MacroCmdLine)

  Set(MacroCmdLine, 'false')
  SetBufferStr(MACRO_NAME + ':ExternalHelpSystem', '', HELP_TOPIC_ID)

  if macro_cmd_line == ''
    configure_helphelp()
  elseif not (macro_cmd_line [1: 2] in '-r', '-h')
    Warn(MACRO_NAME; 'not called with "-r" or "-h"')
  elseif  helphelp_enabled
  and not helphelp_in_progress                // Reject recursive calls.
  and not GetGlobalInt("GETHELP_NoHelpHelp")  // Did GetHelp block helphelp.
    helphelp_in_progress = TRUE
    PushLocation()
    GotoBufferId(HELP_SYSTEMS_ID)
    for i = 1 to NumLines()
      help_system = Trim(GetText(1, MAXSTRINGLEN))
      if help_system <> ''
        if EquiStr(help_system, 'gethelp.mac')
          help_cmd_line = Format(help_system; '-1'; Trim(macro_cmd_line[4:255]))
        else
          help_cmd_line = Format(help_system; macro_cmd_line)
        endif
        if ExecMacro(help_cmd_line)
          if Val(Query(MacroCmdLine))
            Set(MacroCmdLine, 'true')
            SetBufferStr(MACRO_NAME + ':ExternalHelpSystem',
                         SplitPath(help_system, _NAME_),
                         HELP_TOPIC_ID)
            break
          else
            Set(MacroCmdLine, 'false')
          endif
        else
          Warn(MACRO_NAME; 'failed trying to execute macro "', help_system, '".')
          Set(MacroCmdLine, 'false')
        endif
      endif
    endfor
    PopLocation()
    helphelp_in_progress = FALSE
  endif
end

